(* Abstract Syntax Tree and functions for printing it *)

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq |
          And | Or 

type uop = Neg | Not

type typ = Int | Bool | Float | Void | Channel of typ | Array of typ | String
		   | Struct of string

and bind = typ * string

and sdef = StructDef of string * bind list

and expr =
    Literal of int
  | Fliteral of string
  | BoolLit of bool
  | Id of string
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Assign of expr * expr
  | FuncCall of string * expr list
  | GofuncCall of string * expr list
  | NArray of typ * expr
  | ArrAccess of expr * expr
  | Dot of expr * string
  | NChan of typ * expr 
  | Queue of expr * expr
  | StrLit of string
  | NStruct of typ * expr list
  | Noexpr

type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt

type func_decl = {
    isGo: bool;
    typ : typ;
    fname : string;
    formals : bind list;
    locals : bind list;
    body : stmt list;
  }

type program = sdef list * (bind list * func_decl list)

(* Pretty-printing functions *)

let string_of_op = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"

let rec string_of_typ = function
    Int -> "int"
  | Bool -> "bool"
  | Float -> "float"
  | Void -> "void"
  | Channel(ty) -> "channel<" ^ (string_of_typ ty) ^ ">"
  | Array (ty) -> "array<" ^ (string_of_typ ty) ^ ">"
  | String  -> "string"
  | Struct(s) -> s

let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | Fliteral(l) -> l
  | StrLit(l) -> l
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e
  | Assign(v, e) -> string_of_expr v ^ " = " ^ string_of_expr e
  | FuncCall(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | GofuncCall(f, el) ->
     "go " ^ f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Dot(a,n) -> string_of_expr a ^ "." ^ n
  | NArray(ty, e) -> "new(array<" ^ (string_of_typ ty) ^ ">[" ^ (string_of_expr e) ^ "]"
  | ArrAccess(e1, e2) -> string_of_expr e1 ^ "[" ^ string_of_expr e2 ^ "]"
  | NChan(ty, e) -> "new(channel<" ^ string_of_typ ty ^ ">[" ^ string_of_expr e ^ "]"
  | Queue (e1, e2) -> string_of_expr e1 ^ " -> " ^ string_of_expr e2 
  | NStruct (ty,el) -> "new(" ^ string_of_typ ty ^"," ^  String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s

let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let goKey = function
    true -> "gofunction"
  | false -> "function"

let string_of_sdef = function
  StructDef(n,vars) -> "structdef " ^ n ^" {\n"
					   ^ String.concat "" (List.map string_of_vdecl vars)
					   ^"};\n"
    
let string_of_fdecl fdecl =
  goKey fdecl.isGo ^ " " ^
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.locals) ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (sdefs,(vars, funcs)) =
  String.concat "" (List.map string_of_sdef sdefs) ^ "\n" ^
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
