/* Ocamlyacc parser for Go-- */

%{
open Ast
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA PLUS MINUS TIMES DIVIDE ASSIGN /* arithmatic op */
%token NOT EQ NEQ LT LEQ GT GEQ AND OR INC DEC /* logic op */
%token RETURN IF ELSE FOR WHILE /* control flow */
%token INT BOOL FLOAT VOID CHANNEL STRING STRUCT SDEF/* types */
%token LSQUARE RSQUARE NEWCHAN QUEUE /* channel related */
%token NEW ARRAY /* array related */
%token GO FUNCTION GOFUNCTION /* function related */
%token DOT
%token <int> LITERAL
%token <bool> BLIT
%token <string> ID FLIT
%token <string> STRLIT
%token EOF

%start program
%type <Ast.program> program

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%right QUEUE
%right NOT
%right GO
%nonassoc INC DEC
%left DOT

%%

program:
  body EOF { $1 }


body:
  sdef_list decls		{    ($1,$2)       }


sdef_list:
	/* nothing */		{ [] }
 |  sdef sdef_list	   {	$1 :: $2	 }

sdef:
	SDEF ID LBRACE pdecl_opt RBRACE SEMI { StructDef($2,$4)}

decls:
   /* nothing */ { ([], [])               }
 | decls vdecl { (($2 :: fst $1), snd $1) }
 | decls fdecl { (fst $1, ($2 :: snd $1)) }

fdecl:
   isGo typ ID LPAREN formals_opt RPAREN LBRACE vdecl_list stmt_list RBRACE
   {{ 
      isGo = $1; 
      typ = $2;
	    fname = $3;
	    formals = List.rev $5;
	    locals = List.rev $8;
	    body = List.rev $9 } }

isGo:
    FUNCTION   { false }
  | GOFUNCTION { true  }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { $1 }

formal_list:
    typ ID                   { [($1,$2)]     }
  | formal_list COMMA typ ID { ($3,$4) :: $1 }

ptyp:
    INT   { Int   }
  | BOOL  { Bool  }
  | FLOAT { Float }
  | VOID  { Void  }
  | STRING { String }

styp:
  | STRUCT ID { Struct($2) }

typ:
    ptyp	{	$1	}
  | styp    {	$1	}
  | ARRAY LT typ GT { Array($3)}
  | CHANNEL LT typ GT { Channel($3) }

vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
   typ ID SEMI { ($1, $2) }


pdecl_opt:
  /* nothing */			{[]}
  | pdecl_list			{List.rev $1}


pdecl_list:
    pdecl			       { [$1] }
  | pdecl_list pdecl { $2 :: $1 }

pdecl:
   ptyp ID COMMA { ($1, $2) }


stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI                               { Expr $1               }
  | RETURN expr_opt SEMI                    { Return $2             }
  | LBRACE stmt_list RBRACE                 { Block(List.rev $2)    }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7)        }
  | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt
                                            { For($3, $5, $7, $9)   }
  | WHILE LPAREN expr RPAREN stmt           { While($3, $5)         }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
  | expr INC         { Assign($1, Binop($1, Add, Literal(1)))}
  | expr DEC         { Assign($1, Binop($1, Sub, Literal(1)))}
  | expr PLUS   expr { Binop($1, Add,   $3)   }
  | expr MINUS  expr { Binop($1, Sub,   $3)   }
  | expr TIMES  expr { Binop($1, Mult,  $3)   }
  | expr DIVIDE expr { Binop($1, Div,   $3)   }
  | expr EQ     expr { Binop($1, Equal, $3)   }
  | expr NEQ    expr { Binop($1, Neq,   $3)   }
  | expr LT     expr { Binop($1, Less,  $3)   }
  | expr LEQ    expr { Binop($1, Leq,   $3)   }
  | expr GT     expr { Binop($1, Greater, $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3)   }
  | expr AND    expr { Binop($1, And,   $3)   }
  | expr OR     expr { Binop($1, Or,    $3)   }
  | expr QUEUE expr  { Queue($1, $3)   }
  | expr QUEUE       { Queue($1, Noexpr)} /* channel -> */
  | MINUS expr %prec NOT { Unop(Neg, $2)      }
  | NOT expr         { Unop(Not, $2)          }
  | expr ASSIGN expr { Assign($1, $3)         }
  | NEW LPAREN ARRAY LT typ GT LSQUARE expr RSQUARE RPAREN { NArray($5, $8) }
  | NEW LPAREN CHANNEL LT typ GT LSQUARE expr RSQUARE RPAREN { NChan($5, $8) }
  | NEW LPAREN styp COMMA args_opt RPAREN { NStruct($3,$5) }
  | ID LPAREN args_opt RPAREN { FuncCall($1, $3)  }
  | GO ID LPAREN args_opt RPAREN { GofuncCall($2, $4)  }
  | LPAREN expr RPAREN {       $2             }
  | accessor           { $1 }


accessor:
    accessor LSQUARE expr RSQUARE { ArrAccess($1, $3) 	}
  | accessor DOT ID 			  { Dot($1, $3) 		}
  | atom  {$1}


atom: 
    LITERAL          { Literal($1)            }
  | FLIT	           { Fliteral($1)           }
  | BLIT             { BoolLit($1)            }
  | ID               { Id($1)                 }
  | STRLIT           { StrLit($1)             }
    

args_opt:
    /* nothing */ { [] }
  | args_list  { List.rev $1 }

args_list:
    expr                    { [$1] }
  | args_list COMMA expr { $3 :: $1 }
